import argparse
import json
import logging
import os
import re
import sys
from typing import Any, Dict, Optional

import boto3  # type: ignore
from botocore.exceptions import ProfileNotFound  # type: ignore

from iam_ape.aws_iam_actions.scrape_iam_actions import scrape_iam_actions
from iam_ape.evaluator import AuthorizationDetails, EffectivePolicyEvaluator
from iam_ape.helper_functions import deep_update
from iam_ape.helper_types import EntityType

logger = logging.getLogger("IAM-APE")
entity_regex_string = r"arn:aws(-cn|-us-gov)?:iam::\d{12}:(user|group|role)/[\w-]+"
entity_regex = re.compile(entity_regex_string)

banner = """
  __   __   _  _        __   ____  ____ 
 (  ) / _\ ( \/ ) ___  / _\ (  _ \(  __)
  )( /    \/ \/ \(___)/    \ ) __/ ) _) 
 (__)\_/\_/\_)(_/     \_/\_/(__)  (____)
 By Orca Security  https://orca.security
"""


class LogFormatter(logging.Formatter):
    err_fmt = "ERROR: %(msg)s"
    dbg_fmt = "DBG: %(module)s: %(lineno)d: %(msg)s"
    info_fmt = "%(msg)s"

    def __init__(self):
        super().__init__(fmt="%(levelno)d: %(msg)s", datefmt=None, style="%")

    def format(self, record):
        format_orig = self._style._fmt

        if record.levelno == logging.DEBUG:
            self._style._fmt = LogFormatter.dbg_fmt

        elif record.levelno == logging.INFO:
            self._style._fmt = LogFormatter.info_fmt

        elif record.levelno == logging.ERROR:
            self._style._fmt = LogFormatter.err_fmt

        result = logging.Formatter.format(self, record)

        self._style._fmt = format_orig

        return result


def print_banner() -> None:
    logger.info("\033[93m" + banner + "\033[0m")


def initialize_logger() -> None:
    format = LogFormatter()
    handler = logging.StreamHandler(sys.stdout)
    handler.setFormatter(format)
    logging.root.addHandler(handler)
    logging.root.setLevel(logging.INFO)


def validate_arn(arn: str) -> EntityType:
    try:
        assert entity_regex.match(arn)
    except AssertionError:
        raise ValueError(
            f'Invalid ARN format: "{arn}". Expected: "{entity_regex_string}"'
        )

    if ":user/" in arn:
        return EntityType.user
    elif ":group/" in arn:
        return EntityType.group
    else:
        return EntityType.role


def load_auth_details_from_json(inp: str) -> AuthorizationDetails:
    with open(inp, "r") as f:
        auth_report = json.load(f)
    return AuthorizationDetails(auth_report)


def load_auth_details_from_aws(profile: Optional[str] = None) -> AuthorizationDetails:
    logger.info("Attempting to fetch authorization details report from AWS...")
    auth_report: Dict[str, Any] = {}
    profile = profile or os.environ.get("AWS_PROFILE")
    if not profile:
        raise ValueError("No AWS profile found")
    boto3.setup_default_session(profile_name=profile)
    iam_client = boto3.client("iam")
    paginator = iam_client.get_paginator("get_account_authorization_details")
    for page in paginator.paginate():
        auth_report = deep_update(auth_report, page)
    a = AuthorizationDetails(auth_report)
    return a


def build_arg_parser() -> argparse.ArgumentParser:
    arg_parser = argparse.ArgumentParser()
    basic_usage = arg_parser.add_argument_group()
    adv_usage = arg_parser.add_argument_group()
    misc_usage = arg_parser.add_argument_group()

    basic_usage.add_argument(
        "--arn",
        help="ARN of the User/Group/Role for which to evaluate the effective permissions",
    )

    adv_usage.add_argument(
        "-i",
        "--input",
        help='Path to report generated by "aws iam get-account-authorization-details"',
        required=False,
    )
    adv_usage.add_argument(
        "-o",
        "--output",
        help="File path to write the output to (default: stdout)",
        default="stdout",
    )
    adv_usage.add_argument(
        "-f",
        "--format",
        help="Output format",
        choices=["clean", "verbose"],
        default="clean",
    )
    adv_usage.add_argument(
        "-p",
        "--profile",
        help="AWS Profile to use when fetching authorization data",
        required=False,
    )
    adv_usage.add_argument(
        "-v", "--verbose", help="Set verbosity level to DEBUG", action="store_true"
    )

    misc_usage.add_argument(
        "-u",
        "--update",
        help="Update the IAM actions database",
        required=False,
        action="store_true",
    )

    return arg_parser


def main() -> int:
    initialize_logger()
    logger = logging.getLogger("IAM-APE")
    print_banner()

    arg_parser = build_arg_parser()
    arguments = arg_parser.parse_args()

    if arguments.verbose:
        logging.root.setLevel(logging.DEBUG)

    if not (arguments.arn or arguments.update):
        arg_parser.print_help()
        return 0

    if arguments.update:
        scrape_iam_actions()
        if not arguments.arn:
            return 0

    try:
        entity_type = validate_arn(arguments.arn)
    except ValueError as e:
        logger.error(e)
        return -1

    if arguments.input:
        auth_details = load_auth_details_from_json(arguments.input)
    else:
        try:
            auth_details = load_auth_details_from_aws(arguments.profile)
        except ValueError as e:
            logger.error(e)
            return -1
        except ProfileNotFound as e:
            logger.error(e)
            return -1

    logger.info("Evaluating effective permissions")
    calculator = EffectivePolicyEvaluator(auth_details)

    try:
        res = calculator.evaluate(arn=arguments.arn, entity_type=entity_type)
    except ValueError as e:
        logger.error(e)
        return -1

    if arguments.format == "clean":
        out = calculator.policy_expander.shrink_policy(res.allowed_permissions)
    else:
        out = res.to_dict()

    if arguments.output == "stdout":
        logger.info(f"Effective permissions policy for {arguments.arn}\n")
        logger.info(json.dumps(out, indent=2))
    else:
        with open(arguments.output, "w") as f:
            json.dump(out, f)

    return 0


if __name__ == "__main__":
    exit(main())



import json
import logging
import tarfile
import tempfile
from collections import defaultdict
from typing import Any, Dict

import requests
import tqdm  # type: ignore
from bs4 import BeautifulSoup  # type: ignore

from iam_ape.consts import actions_json_location

logger = logging.getLogger("IAM-APE:updater")
base_url = "https://docs.aws.amazon.com/service-authorization/latest/reference/"


def get_soup(url: str) -> BeautifulSoup:
    html_doc = requests.get(url).content.decode("utf-8")
    return BeautifulSoup(html_doc, "html.parser")


def scrape_iam_actions() -> int:
    soup = get_soup(base_url + "reference_policies_actions-resources-contextkeys.html")
    all_a = soup.find_all("a")
    all_links = [
        a.get("href") for a in all_a if a.get("href", "").startswith("./list_")
    ]
    data: Dict[str, Any] = {}

    logger.info("Updating AWS IAM actions database...")
    for link in tqdm.tqdm(all_links, ncols=70):
        try:
            soup = get_soup(base_url + link[2:])
            service_prefix = soup.find("code").string
            data[service_prefix] = defaultdict(dict)
            table = soup.find("table")
            table_rows = table.find_all("tr")
            for row in table_rows[1:]:
                all_cells = row.find_all("td")

                try:
                    action = all_cells[0].find("a", href=True).get_text()
                except AttributeError:  # AWS doing AWS things
                    continue

                if desc := all_cells[1].string:
                    data[service_prefix][action]["description"] = desc
                else:
                    continue

                data[service_prefix][action]["access"] = all_cells[2].string

        except AttributeError as e:
            logger.error(f"Error occurred while processing {link} - {e}")
            raise e

    logger.info("Done!")

    with tempfile.NamedTemporaryFile("w+") as f:
        json.dump(data, f, indent=2)
        f.flush()

        with tarfile.open(actions_json_location, "w:gz") as tar:
            tar.add(f.name, arcname="actions.json")

    return 0


if __name__ == "__main__":
    exit(scrape_iam_actions())



import json
import logging
import tarfile
from collections import defaultdict
from fnmatch import fnmatch
from typing import Any, Dict, FrozenSet, List, Literal, Optional, Set, Tuple

from iam_ape.consts import PolicyElement, actions_json_location
from iam_ape.helper_functions import as_list, normalize_policy
from iam_ape.helper_types import (
    Action,
    AwsPolicyStatementType,
    AwsPolicyType,
    HashableDict,
    PermissionsContainer,
    PolicyWithSource,
)

logger = logging.getLogger("policy expander")


def _append_action(
    res: Dict[str, Set[Action]],
    action: str,
    resources: Optional[List[str]],
    not_resources: Optional[List[str]],
    condition: Optional[Dict[str, Any]],
    source: str,
) -> None:
    for resource in resources or []:
        if condition:
            for key, val in condition.items():
                res[action].add(
                    Action(
                        action=action,
                        resource=resource,
                        not_resource=None,
                        condition=HashableDict.recursively({key: val}),
                        source=source,
                    )
                )
        else:
            res[action].add(
                Action(
                    action=action,
                    resource=resource,
                    not_resource=None,
                    condition=None,
                    source=source,
                )
            )
    for not_resource in not_resources or []:
        if condition:
            for key, val in condition.items():
                res[action].add(
                    Action(
                        action=action,
                        resource=None,
                        not_resource=not_resource,
                        condition=HashableDict.recursively({key: val}),
                        source=source,
                    )
                )
        else:
            res[action].add(
                Action(
                    action=action,
                    resource=None,
                    not_resource=not_resource,
                    condition=None,
                    source=source,
                )
            )


class PolicyExpander:
    def __init__(self) -> None:
        self.all_iam_actions: Dict[str, Any] = self._init_iam_actions()
        self._all_service_wildcards: List[str] = [
            f"{k}:*" for k, v in self.all_iam_actions.items() if len(v) > 0
        ]

    @staticmethod
    def _init_iam_actions() -> Dict[str, Dict[str, Dict[str, str]]]:
        try:
            with tarfile.open(actions_json_location) as f:
                return json.load(f.extractfile("actions.json"))  # type: ignore
        except Exception as e:
            logger.exception(f"Failed to load AWS IAM Actions due to {e}")
        return {}

    def expand_policies(self, policies: List[PolicyWithSource]) -> PermissionsContainer:
        allow_actions_dict: Dict[str, Set[Action]] = defaultdict(set)
        deny_actions_dict: Dict[str, Set[Action]] = defaultdict(set)

        all_actions: Dict[Literal["Allow", "Deny"], Dict[str, Set[Action]]] = {
            PolicyElement.ALLOW: allow_actions_dict,
            PolicyElement.DENY: deny_actions_dict,
        }

        for source, policy in policies:
            if policy is None:
                continue
            for i, statement in enumerate(policy["Statement"]):
                sid = "{}/{}".format(source, statement.get("Sid", i))
                if actions := statement.get(PolicyElement.ACTION):
                    for action in actions:
                        expanded_action = self.expand_action(
                            Action(
                                action=action,
                                resource=statement.get(PolicyElement.RESOURCE),
                                not_resource=statement.get(PolicyElement.NOTRESOURCE),
                                condition=statement.get(PolicyElement.CONDITION),
                                source=sid,
                            )
                        )
                        for action_key, action_values in expanded_action.items():
                            all_actions[statement[PolicyElement.EFFECT]][
                                action_key
                            ].update(action_values)

                elif statement.get(PolicyElement.NOTACTION):
                    expanded_notaction = self.expand_not_action(statement, sid)
                    for action_key, action_values in expanded_notaction.items():
                        all_actions[statement[PolicyElement.EFFECT]][action_key].update(
                            action_values
                        )

        return PermissionsContainer(
            allowed_permissions=all_actions[PolicyElement.ALLOW],
            denied_permissions=all_actions[PolicyElement.DENY],
        )

    def expand_action(self, iam_action: Action) -> Dict[str, Set[Action]]:
        res: Dict[str, Set[Action]] = defaultdict(set)

        if iam_action.action == PolicyElement.WILDCARD:  # {"Action": ["*"]}
            for service, action_dicts in self.all_iam_actions.items():
                for action in action_dicts.keys():
                    _append_action(
                        res=res,
                        action=f"{service}:{action}",
                        resources=as_list(iam_action.resource),
                        not_resources=as_list(iam_action.not_resource),
                        condition=iam_action.condition,
                        source=iam_action.source,
                    )
        elif PolicyElement.WILDCARD in iam_action.action:  # {"Action": ["iam:*"]}
            assert (
                iam_action.action.count(":") == 1
            ), f"Got an invalid wildcard action: {iam_action.action}"
            service, wildcard_action = iam_action.action.split(":")
            for action in self.all_iam_actions[service].keys():
                if fnmatch(action, wildcard_action):
                    _append_action(
                        res=res,
                        action=f"{service}:{action}",
                        resources=as_list(iam_action.resource),
                        not_resources=as_list(iam_action.not_resource),
                        condition=iam_action.condition,
                        source=iam_action.source,
                    )
        else:  # {"Action": ["sts:GetCallerIdentity"]}
            _append_action(
                res=res,
                action=iam_action.action,
                resources=as_list(iam_action.resource),
                not_resources=as_list(iam_action.not_resource),
                condition=iam_action.condition,
                source=iam_action.source,
            )

        return res

    def expand_not_action(
        self, statement: AwsPolicyStatementType, sid: str
    ) -> Dict[str, Set[Action]]:
        res: Dict[str, Set[Action]] = defaultdict(set)
        notactions: List[str] = statement.get(PolicyElement.NOTACTION) or []
        if any(
            [notaction == PolicyElement.WILDCARD for notaction in notactions]
        ):  # {"NotAction": ["*"]}
            # This is here as a safeguard. No sane person should write a policy like this. It has no effect.
            pass
        else:  # {"NotAction": ["ec2:*", "iam:Get*", "sts:GetCallerIdentity"]}
            for iam_service, action_dicts in self.all_iam_actions.items():
                for action in action_dicts.keys():
                    curr_action = f"{iam_service}:{action}"
                    if any(
                        [fnmatch(curr_action, not_action) for not_action in notactions]
                    ):
                        continue
                    _append_action(
                        res=res,
                        action=curr_action,
                        resources=statement.get(PolicyElement.RESOURCE),
                        not_resources=statement.get(PolicyElement.NOTRESOURCE),
                        condition=statement.get(PolicyElement.CONDITION),
                        source=sid,
                    )

        return res

    def deflate_policy_statements(
        self,
        policy_statements: List[AwsPolicyStatementType],
    ) -> List[AwsPolicyStatementType]:
        for statement in policy_statements:
            action_dict = defaultdict(list)
            for action in statement.get(PolicyElement.ACTION) or []:
                service, action_key = action.split(":")
                action_dict[service].append(action_key)
            for service, action_keys in action_dict.items():
                try:
                    if all(
                        [
                            action in action_keys
                            for action in self.all_iam_actions[service]
                        ]
                    ):
                        for action_key in action_keys:
                            statement[PolicyElement.ACTION].remove(f"{service}:{action_key}")  # type: ignore
                        statement[PolicyElement.ACTION].append(f"{service}:*")  # type: ignore
                except KeyError as e:
                    logger.exception(f"Unknown service: {service=}")
                    raise e
            if all(
                [
                    action in as_list(statement[PolicyElement.ACTION])
                    for action in self._all_service_wildcards
                ]
            ):
                statement[PolicyElement.ACTION] = ["*"]
        return policy_statements

    def shrink_policy(self, allow_actions: Dict[str, Set[Action]]) -> AwsPolicyType:
        policy_res: AwsPolicyType = {"Statement": []}
        squashed_policies: Set[Action] = set()
        for allow_action_set in allow_actions.values():
            squashed_policies.update(allow_action_set)

        by_resource_notresource: Dict[
            Tuple[Optional[str], Optional[str]], Dict[Optional[HashableDict], Set[str]]
        ] = defaultdict(lambda: defaultdict(set))
        for action_tuple in squashed_policies:
            by_resource_notresource[(action_tuple.resource, action_tuple.not_resource)][
                HashableDict.recursively(action_tuple.condition)
            ].add(action_tuple.action)

        by_action_condition_notresource: Dict[
            Tuple[FrozenSet[str], Optional[HashableDict], Optional[str]], Set[str]
        ] = defaultdict(set)
        for resource_notresource, condition_actions in by_resource_notresource.items():
            resource, notresource = resource_notresource
            for condition, actions_set in condition_actions.items():
                key = frozenset(actions_set), condition, notresource
                if resource:
                    by_action_condition_notresource[key].add(resource)
                elif not by_action_condition_notresource.get(key):
                    by_action_condition_notresource[key] = set()

        statements = list()
        for (
            action_condition_notresource,
            resources,
        ) in by_action_condition_notresource.items():
            actions, condition, notresource = action_condition_notresource
            statement: AwsPolicyStatementType = {
                "Effect": "Allow",
                PolicyElement.ACTION: sorted(list(actions)),
            }
            if resources:
                statement[PolicyElement.RESOURCE] = sorted([r for r in resources])
            if notresource:
                statement[PolicyElement.NOTRESOURCE] = as_list(notresource)
            if condition:
                statement[PolicyElement.CONDITION] = condition
            statements.append(statement)

        policy_res["Statement"] = self.deflate_policy_statements(statements)
        return normalize_policy(policy_res)
